import asyncio
import logging
from typing import Literal

import asyncssh

from kalandra.auth.basic import CredentialProvider
from kalandra.gitprotocol import PacketLine

from .base import (
    BaseConnection,
    ConnectionException,
    FetchConnection,
    PushConnection,
    Transport,
)

logger = logging.getLogger(__name__)


class SSHConnection(BaseConnection["SSHTransport"]):
    _ssh: asyncssh.SSHClientConnection

    async def _open_service_connection(
        self, service_name: Literal["git-upload-pack", "git-receive-pack"]
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        logger.debug("Connecting to %s", self.transport.url)

        options = asyncssh.SSHClientConnectionOptions(
            # ignore_encrypted=False,
            # passphrase=(lambda x: input("Enter passphrase for %s: " % x)), # type: ignore
        )
        # open the SSH connection
        self._ssh = await asyncssh.connect(
            host=self.transport.host,
            username=self.transport.user,
            port=self.transport.port,
            options=options,
        )

        try:
            # run the git-upload-pack command
            self._process = await self._ssh.create_process(
                f"{service_name} {self.transport.path}",
                env={"GIT_PROTOCOL": f"version={self.git_protocol}"},
                encoding=None,
            )

            return self._process.stdout, self._process.stdin  # type: ignore
        except Exception:
            self._ssh.close()
            raise

    async def _close_service_connection(self) -> None:
        try:
            self._process.stdin.write_eof()
            await self._process.wait()
        finally:
            self._ssh.close()

    async def _read_packet(self) -> PacketLine:
        try:
            if self._process.exit_status is not None:
                raise EOFError("Process has exited")
            return await super()._read_packet()
        except asyncio.IncompleteReadError:
            logger.debug("EOF during packet read", exc_info=True)
            errors = (await self._process.stderr.read()).decode("utf-8")
            raise ConnectionException("Server closed connection: " + errors) from None


class SSHFetchConnection(SSHConnection, FetchConnection["SSHTransport"]):
    async def _open_fetch_service_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-upload-pack")


class SSHPushConnection(SSHConnection, PushConnection["SSHTransport"]):
    async def _open_push_service_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-receive-pack")


class SSHTransport(Transport):
    def __init__(self, *, url: str, credentials_provider: CredentialProvider):
        assert url.startswith("ssh://")
        super().__init__(url=url, credentials_provider=credentials_provider)

        address, self.path = url[6:].split("/", 1)
        self.user, host = address.split("@", 1)

        logger.debug("Resolved SSH host: %s", host)

        if ":" in host:
            self.host, port = host.split(":", 1)
            self.port = int(port)
        else:
            self.host = host
            self.port = 22

    @classmethod
    def can_handle_url(cls, url: str) -> bool:
        return url.startswith("ssh://")

    def fetch(self) -> SSHFetchConnection:
        return SSHFetchConnection(transport=self)

    def push(self) -> SSHPushConnection:
        return SSHPushConnection(transport=self)

    @property
    def url(self) -> str:
        return f"ssh://{self.user}@{self.host}:{self.port}/{self.path}"



from .base import BaseConnection, FetchConnection, Transport

# Import all the transports so they can be used
from .file import FileTransport
from .http import HTTPTransport
from .ssh import SSHTransport

__all__ = [
    "BaseConnection",
    "FetchConnection",
    "FileTransport",
    "SSHTransport",
    "HTTPTransport",
    "Transport",
]



"""
Tools related to git protocol v2.

See: https://git-scm.com/docs/protocol-v2

"""

from collections.abc import Buffer
from enum import Enum
from typing import NamedTuple


class PacketLineType(Enum):
    DATA = -1
    FLUSH = 0
    DELIMITER = 1
    RESPONSE_END = 2
    UNKNOWN = 3

    @classmethod
    def from_bytes(cls, data: bytes) -> "tuple[PacketLineType, int]":
        if len(data) != 4:
            raise ValueError("Invalid packet line marker size: %r" % len(data))
        pkt_marker = int(data, 16)
        if pkt_marker >= 4:
            return cls.DATA, pkt_marker - 4
        else:
            return cls(pkt_marker), 0


class PacketLine:
    """
    Represents a single packet line in the git protocol.

    See: https://git-scm.com/docs/gitprotocol-common#_pkt_line_format
    """

    __slots__ = ("length", "data", "type")

    length: int
    data: bytes
    type: PacketLineType

    FLUSH: "PacketLine"
    DELIMITER: "PacketLine"

    @classmethod
    def from_buffer(cls, data: Buffer) -> "PacketLine":
        pkt_type, pkt_payload_length, pkt_payload = cls.sniff_buffer(data)
        if pkt_type is None:
            raise ValueError("Not enough data to determine packet type")
        if pkt_payload_length < 0:
            raise ValueError("Not enough data to read the whole packet, need %d more bytes" % -pkt_payload_length)

        return cls.from_marker_and_payload(pkt_type, pkt_payload)

    @classmethod
    def from_marker_and_payload(cls, pkt_type: PacketLineType, payload: Buffer | None) -> "PacketLine":
        if pkt_type != PacketLineType.DATA:
            return cls(0, b"", pkt_type)
        else:
            if payload is None:
                raise ValueError("Payload is required for DATA packet type")
            mview = memoryview(payload)
            return cls(len(mview), bytes(mview), pkt_type)

    @classmethod
    def data_from_string(cls, data: str) -> "PacketLine":
        encoded = f"{data}\n".encode("utf-8")
        return cls(len(encoded), encoded, PacketLineType.DATA)

    @classmethod
    def sniff_buffer(cls, data: Buffer) -> "tuple[PacketLineType | None, int, Buffer | None]":
        """
        Sniff the buffer to determine the type of packet line and its length.

        If there is not enough data to determine the type, returns None and a negative number
        indicating the number of bytes missing to determine the type.

        If there is enough data to determine the type, but not enough to read the whole payload,
        returns the type and a negative number indicating the number of bytes missing to read the whole payload.

        If there is enough data to determine the type and read the whole packet,
        returns the type and the length of the packet's payload.
        """
        mview = memoryview(data)
        mview_length = len(mview)
        if mview_length < 4:
            # Not enough data to read the length
            return None, 4 - mview_length, None

        pkt_marker = int(bytes(mview[:4]), 16)
        if pkt_marker >= 4:
            if pkt_marker > len(mview):
                # we know the length, but we don't have enough data to read the whole packet
                return PacketLineType.DATA, len(mview) - pkt_marker, None
            return PacketLineType.DATA, pkt_marker - 4, mview[4:pkt_marker]
        else:
            return PacketLineType(pkt_marker), 0, None

    def __init__(self, length: int, data: bytes, type: PacketLineType):
        self.length = length
        self.data = data
        self.type = type

    @property
    def marker_bytes(self) -> bytes:
        return b"%04x" % (self.length + 4 if self.type == PacketLineType.DATA else self.type.value)

    @property
    def data_decoded(self) -> str:
        return self.data.decode("utf-8").rstrip()

    def __repr__(self) -> str:
        if self.length < 100:
            return f"PacketLine({self.length}, {self.data!r}, {self.type})"
        else:
            return f"PacketLine({self.length}, {self.data[:100]!r}..., {self.type})"


PacketLine.FLUSH = PacketLine.from_marker_and_payload(PacketLineType.FLUSH, None)
PacketLine.DELIMITER = PacketLine.from_marker_and_payload(PacketLineType.DELIMITER, None)


NULL_OBJECT_ID = "0" * 40


class Ref(NamedTuple):
    name: str
    object_id: str

    @classmethod
    def from_line(cls, line: str) -> "Ref":
        parts = line.split(" ")
        if len(parts) != 2:
            raise ValueError(f"Invalid ref line: {line}")
        return cls(name=parts[1], object_id=parts[0])


class RefChange(NamedTuple):
    ref: str
    old: str
    new: str

    def __str__(self) -> str:
        if self.is_create:
            return f"CREATE {repr(self.ref)} {self.new}"
        elif self.is_delete:
            return f"DELETE {repr(self.ref)} {self.old}"
        else:
            return f"UPDATE {repr(self.ref)} {self.old}..{self.new}"

    @property
    def is_delete(self) -> bool:
        return self.new == NULL_OBJECT_ID

    @property
    def is_create(self) -> bool:
        return self.old == NULL_OBJECT_ID

    @property
    def is_update(self) -> bool:
        return self.old != NULL_OBJECT_ID and self.new != NULL_OBJECT_ID



import asyncio
import logging
import os
import time
from typing import AsyncIterator, Literal
from urllib.parse import urlparse, urlunsplit

import aiohttp
from aiofiles.threadpool.binary import AsyncBufferedIOBase

from kalandra.auth.basic import CredentialProvider
from kalandra.gitprotocol import PacketLine, PacketLineType

from .base import (
    BaseConnection,
    ConnectionException,
    FetchConnection,
    PushConnection,
    Transport,
)

logger = logging.getLogger(__name__)

type SessionFactory = type[aiohttp.ClientSession]


def http_timeout():
    total_timeout = int(os.environ.get("KALANDRA_HTTP_TIMEOUT", 1200))  # default 20 minutes
    logger.debug("HTTP timeout set to %d seconds", total_timeout)
    return aiohttp.ClientTimeout(total=total_timeout, connect=60, sock_connect=60)


class HTTPSmartConnection(BaseConnection["HTTPTransport"]):
    _session: aiohttp.ClientSession | None = None
    _service: str | None = None

    async def _open_service_connection(
        self, service_name: Literal["git-upload-pack", "git-receive-pack"]
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        origin = urlparse(self.transport.url).hostname
        assert origin is not None, "No hostname in service URL"
        logger.debug("Getting credentials for %s", origin)
        credentials = await self.transport.get_credentials(origin)

        self._session = self.transport.session_factory(
            headers={
                "User-Agent": "git/2.46.0",
            },
            auth=aiohttp.BasicAuth(*credentials) if credentials else None,
        )

        # As per https://git-scm.com/docs/gitprotocol-http#_url_format
        service_url = self.transport.url + f"/info/refs?service={service_name}"

        logger.debug("Connecting to %s, protocol %s", service_url, self.git_protocol)
        hello_resp = await self._session.get(
            service_url,
            headers={
                "Accept": f"application/x-{service_name}-advertisement",
                "Git-Protocol": f"version={self.git_protocol}",
            },
        )

        if hello_resp.status != 200:
            raise ConnectionException(f"Failed to connect to server {hello_resp.reason} ({hello_resp.status})")

        content_type = hello_resp.headers.get("Content-Type")
        if content_type != f"application/x-{service_name}-advertisement":
            raise ConnectionException(f"Expected 'smart' HTTP response, got {content_type}")

        # Read the first packet and verify it is the service we requested
        self._service = service_name
        self.reader = hello_resp.content  # type: ignore

        header = await self._read_packet()

        if header.type != PacketLineType.DATA:
            raise ConnectionException(f"Expected data packet, got {header}")

        if header.data == f"# service={service_name}\n".encode("ascii"):
            # The server sent the service name. GitHub does this even for v2, so we need to read the next packet
            pkt = await self._read_packet()
            assert pkt.type == PacketLineType.FLUSH, "Expected flush after service header, got %r" % pkt

            pkt = await self._read_packet()
            if pkt.type == PacketLineType.DATA and pkt.data.rstrip() == b"version 2":
                logger.debug("Server supports protocol version 2 (with quirks)")
                self._negotiated_protocol = 2
            else:
                self._negotiated_protocol = 1
                logger.debug("Server supports protocol version 1")

            # put the packet back in the buffer
            self._shift_packet(pkt)

        elif header.data.rstrip() == b"version 2":
            logger.debug("Server supports protocol version 2")
            self._negotiated_protocol = 2
            self._shift_packet(header)
        else:
            raise ConnectionException(f"Expected service header or version packet, instead got: {header}")

        return hello_resp.content, None  # type: ignore

    async def _close_service_connection(self) -> None:
        if self._session:
            await self._session.close()


class HTTPSmartFetchConnection(HTTPSmartConnection, FetchConnection["HTTPTransport"]):
    async def _open_fetch_service_connection(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-upload-pack")

    async def _send_packet_transaction(self, packets: AsyncIterator[PacketLine]) -> None:
        assert self.writer is None
        assert self._session is not None

        async def generate_command_data() -> AsyncIterator[bytes]:
            async for pkt in packets:
                yield pkt.marker_bytes
                yield pkt.data

        # Send a new HTTP POST request with the command
        url = self.transport.url + f"/{self._service}"
        logger.debug("Sending 'fetch' request %s", url)
        response = await self._session.post(
            url,
            headers={
                "Content-Type": f"application/x-{self._service}-request",
                "Cache-Control": "no-cache",
                "Accept": f"application/x-{self._service}-result, */*",
                "Git-Protocol": f"version={self.negotiated_protocol}",
            },
            data=generate_command_data(),
            timeout=http_timeout(),
        )

        if response.status != 200:
            logger.error("Error response: %s", await response.text())
            raise ConnectionException(f"Failed to send packets: {response.reason} ({response.status})")

        logger.debug("Request complete, got: %s", response.headers)

        self.reader = response.content  # type: ignore


MEGABTE = 1024 * 1024


class HTTPSmartPushConnection(HTTPSmartConnection, PushConnection["HTTPTransport"]):
    async def _open_push_service_connection(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-receive-pack")

    async def _send_commands(self, packets: AsyncIterator[PacketLine], packfile: AsyncBufferedIOBase | None) -> None:
        """
        As per <https://git-scm.com/docs/http-protocol#_smart_service_git_receive_pack>, the client must
        make a new POST request directly to the service URL.
        """
        assert self._session is not None, "HTTP session not initialized"
        assert self.writer is None

        async def generate_command_data():
            async for pkt in packets:
                yield pkt.marker_bytes
                yield pkt.data

            if packfile is not None:
                logger.debug("Sending packfile")
                await packfile.seek(0)
                bytes_sent = 0
                total_sent = 0
                start_time = time.perf_counter()
                async for chunk in packfile:
                    yield chunk
                    bytes_sent += len(chunk)
                    if bytes_sent > 32 * MEGABTE:
                        end_time = time.perf_counter()
                        total_sent += bytes_sent
                        elapsed_seconds = end_time - start_time
                        logger.debug(
                            "Sent %.2f MB so far, current speed: %.2f MB/s",
                            total_sent / MEGABTE,
                            bytes_sent / (elapsed_seconds) / MEGABTE,
                        )
                        bytes_sent = 0
                        start_time = time.perf_counter()

                logger.debug("Packfile sent.")
            else:
                logger.debug("No packfile to send")

        # Send a new HTTP POST request with the command
        url = self.transport.url + f"/{self._service}"
        logger.debug("Sending PUSH command to %s", url)
        response = await self._session.post(
            url,
            headers={
                "Content-Type": f"application/x-{self._service}-request",
                "Cache-Control": "no-cache",
                "Accept": f"application/x-{self._service}-result, */*",
                "Git-Protocol": f"version={self.negotiated_protocol}",
            },
            timeout=http_timeout(),
            data=generate_command_data(),
        )
        if response.status != 200:
            logger.error("Error response: %s", await response.text())
            raise ConnectionException(f"Request failed: {response.status}: {response.reason}")

        self.reader = response.content  # type: ignore


class HTTPTransport(Transport):
    def __init__(
        self,
        *,
        url: str,
        credentials_provider: CredentialProvider,
        session_factory: SessionFactory = aiohttp.ClientSession,
    ):
        super().__init__(url=url, credentials_provider=credentials_provider)

        parsed = urlparse(url)
        assert parsed.scheme in ("http", "https"), f"Unsupported protocol: {parsed.scheme}"
        self.protocol = parsed.scheme

        self.user = parsed.username
        self.host = parsed.hostname
        self.port = parsed.port
        self.path = parsed.path
        self._session_factory = session_factory

    @classmethod
    def can_handle_url(cls, url: str) -> bool:
        return url.startswith(("https://", "http://"))

    def fetch(self) -> HTTPSmartFetchConnection:
        return HTTPSmartFetchConnection(transport=self)

    def push(self) -> HTTPSmartPushConnection:
        return HTTPSmartPushConnection(transport=self)

    @property
    def url(self) -> str:
        netloc = f"{self.host}:{self.port}" if self.port else self.host
        return urlunsplit((self.protocol, netloc, self.path, "", ""))

    @property
    def session_factory(self) -> SessionFactory:
        return self._session_factory



from .basic import ChainedCredentialProvider, CredentialProvider, NetrcCredentialProvider, NoopCredentialProvider

__all__ = [
    "CredentialProvider",
    "NetrcCredentialProvider",
    "NoopCredentialProvider",
    "ChainedCredentialProvider",
]



import argparse
import logging
import pathlib

from .auth import ChainedCredentialProvider, NetrcCredentialProvider
from .commands.update_mirror import create_glob_filter, update_mirror
from .transports import Transport

logger = logging.getLogger(__name__)


def create_parser():
    parser = argparse.ArgumentParser(
        description="Update a mirror of a git repository",
        exit_on_error=False,
    )

    parser.add_argument(
        "--source",
        help="URL of the repository to take changes from",
        required=True,
    )
    parser.add_argument(
        "--target",
        help="URL of the repository to push changes to",
        required=True,
    )

    parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="Don't download or push any changes, just print the ref differences",
    )

    parser.add_argument(
        "--netrc",
        help="Path to the .netrc file to read credentials from. If not provided, the default location is used.",
        type=pathlib.Path,
        nargs="?",
        const=True,
    )

    parser.add_argument(
        "--github-app-id",
        help="GitHub Application ID to be used for HTTP authentication.",
    )

    parser.add_argument(
        "--github-app-key",
        help="GitHub Application Private Key to be used for HTTP authentication.",
        type=pathlib.Path,
    )

    parser.add_argument(
        "--github-org",
        help="GitHub Org to use for GitHub App authentication",
    )

    parser.add_argument(
        "--include-ref",
        help="Include only refs that match the given glob patter. Default: refs/heads/*, refs/tags/*",
        action="append",
        default=["refs/heads/*", "refs/tags/*"],
    )

    parser.add_argument(
        "--exclude-ref",
        help="Exclude refs that match the given glob pattern. This takes precedence over --include-ref",
        action="append",
        default=[],
    )

    parser.add_argument(
        "--log-level",
        help="Set the logging level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )

    return parser


async def main(cmdline_args: list[str]) -> int:
    parser = create_parser()
    try:
        args = parser.parse_args(cmdline_args)
    except argparse.ArgumentError as e:
        logger.error("Error parsing arguments: %s", e)
        parser.print_help()
        return 1

    logging.root.setLevel(args.log_level)

    include_filter = create_glob_filter(*args.include_ref)
    exclude_filter = create_glob_filter(*args.exclude_ref)

    # Credentials provider
    credentials_provider = ChainedCredentialProvider()
    if args.netrc is not None:
        netrc_file = args.netrc if isinstance(args.netrc, pathlib.Path) else None
        credentials_provider.add_provider(NetrcCredentialProvider(netrc_file))

    if args.github_app_id or args.github_app_key:
        assert args.github_app_id and args.github_app_key, "GitHub App requires both ID and private key"
        try:
            from kalandra.github_utils import GithubAPI
        except ImportError:
            logger.error("GitHub integration not available, install kalandra[github] to enable it.")
            return 1

        github_api = GithubAPI(args.github_app_id, args.github_app_key)

        credentials_provider.add_provider(github_api.crendentials_provider_for_org(args.github_org))
    else:
        github_api = None

    source_url = args.source

    if source_url.startswith("target-prop:"):
        if github_api is None:
            logger.error("GitHub App credentials are required to use target-prop")
            return 1

        logger.info("Looking up source URL from target repository")

        props = await github_api.get_repo_properties(args.target)
        source_prop_name = args.source[len("target-prop:") :]

        if source_prop_name not in props:
            logger.error("Property %s not found in target repository", args.source)
            return 1

        source_prop_value = props[source_prop_name]
        if source_prop_value.startswith("/"):
            # Assume it's a relative path and lookup host from another prop
            if f"{source_prop_name}-host" not in props:
                logger.error(
                    "Source property is relative, but no %s-host not found in target repository: %s",
                    args.source,
                    source_prop_value,
                )
                return 1
            source_url = props[f"{source_prop_name}-host"] + source_prop_value
        else:
            source_url = source_prop_value

    source = Transport.from_url(source_url, credentials_provider=credentials_provider)
    target = Transport.from_url(args.target, credentials_provider=credentials_provider)

    logger.info("Update start [%s -> %s]", source.url, target.url)
    try:
        await update_mirror(
            source,
            target,
            dry_run=args.dry_run,
            include_filter=include_filter,
            exclude_filter=exclude_filter,
        )
        logger.info("Update success [%s -> %s]", source.url, target.url)
        return 0
    except Exception as e:
        logger.error("Update failed [%s -> %s]", source.url, target.url, exc_info=e)
        return 1



import asyncio
import logging
import sys

from kalandra.cli import main


def run():
    logging.basicConfig(
        level=logging.INFO,
        format="[%(asctime)s] %(levelname)-8s | %(filename)s:%(lineno)s | %(message)s",
        datefmt="%H:%M:%S",
    )
    result = asyncio.run(main(sys.argv[1:]))
    sys.exit(result)


if __name__ == "__main__":
    run()



import itertools
import logging
from abc import ABCMeta, abstractmethod
from asyncio import IncompleteReadError, StreamReader, StreamWriter
from typing import AsyncIterator, Iterable

from aiofiles.threadpool.binary import AsyncBufferedIOBase

from kalandra.auth.basic import CredentialProvider
from kalandra.gitprotocol import NULL_OBJECT_ID, PacketLine, PacketLineType, Ref, RefChange

logger = logging.getLogger(__name__)

MEGABYTES = 1024 * 1024


class ConnectionException(Exception):
    pass


class ServerError(Exception):
    """
    The server returned an error response.
    """

    pass


class Transport(metaclass=ABCMeta):
    def __init__(self, *, url: str, credentials_provider: CredentialProvider):
        self._url = url
        self.credentials_provider = credentials_provider

    @abstractmethod
    def fetch[T: Transport](self: T) -> "FetchConnection[T]":
        """
        Initialize the connection to the remote service (git-upload-pack or git-receive-pack).
        """
        pass

    @abstractmethod
    def push[T: Transport](self: T) -> "PushConnection[T]":
        """
        Initialize the connection to the remote service (git-upload-pack or git-receive-pack).
        """
        pass

    @property
    def url(self) -> str:
        return self._url

    async def get_credentials(self, origin: str) -> tuple[str, str] | None:
        return await self.credentials_provider.get_credentials(origin)

    @classmethod
    @abstractmethod
    def can_handle_url(cls, url: str) -> bool:
        """
        Check if this transport can handle the given URL.
        """
        return False

    @classmethod
    def from_url(cls, url: str, *, credentials_provider: CredentialProvider) -> "Transport":
        """
        Create a transport instance from a URL.
        """
        assert credentials_provider is not None, "Credentials provider is required"
        for cls in cls.__subclasses__():
            if cls.can_handle_url(url):
                return cls(url=url, credentials_provider=credentials_provider)
        raise ValueError(f"Unsupported URL: {url}")


class BaseConnection[T: Transport]:
    capabilities: frozenset[str]
    transport: T
    reader: StreamReader | None
    writer: StreamWriter | None
    git_protocol: int | None

    def __init__(self, *, transport: T):
        self.transport = transport
        self.capabilities = frozenset()
        self.reader = None
        self.writer = None

        self.git_protocol = None
        self._negotiated_protocol = None
        self.last_packet: PacketLine | None = None

        self._shifted_packets: list[PacketLine] = []

    @property
    def negotiated_protocol(self) -> int | None:
        if self._negotiated_protocol is not None:
            return self._negotiated_protocol
        return self.git_protocol

    @abstractmethod
    async def _close_service_connection(self) -> None:
        """
        Close the connection to the remote service.
        """
        pass

    async def _read_packet(self) -> PacketLine:
        assert self.reader is not None

        if self._shifted_packets:
            # There are packets that were shifted, return the first one
            return self._shifted_packets.pop(0)

        pkt_marker = await self.reader.readexactly(4)
        pkt_type, payload_length = PacketLineType.from_bytes(pkt_marker)

        if pkt_type == PacketLineType.DATA:
            pkt_payload = await self.reader.readexactly(payload_length)
            return PacketLine.from_marker_and_payload(pkt_type, pkt_payload)
        else:
            return PacketLine.from_marker_and_payload(pkt_type, None)

    def _at_eof(self):
        return not self._shifted_packets and (self.reader is None or self.reader.at_eof())

    def _shift_packet(self, packet: PacketLine) -> None:
        self._shifted_packets.append(packet)

    async def _read_packets_until_flush(self) -> AsyncIterator[PacketLine]:
        assert self.reader is not None

        while not self._at_eof():
            try:
                packet = await self._read_packet()
                if packet.type == PacketLineType.FLUSH:
                    self.last_packet = packet
                    return
                yield packet
            except IncompleteReadError as e:
                if len(e.partial) > 0:
                    raise ValueError(f"Unexpected EOF while reading packet length: {e.partial}")
                raise ValueError("Reached EOF before FLUSH packet") from e

    async def _read_packets_section(self) -> AsyncIterator[PacketLine]:
        """
        Read packets until a either delimiter or flush packet is received.

        The caller can check which packet was received by looking at the last_packet attribute after the iteration is done.
        """
        self.last_packet = None
        while True:
            try:
                packet = await self._read_packet()
                if packet.type in (PacketLineType.FLUSH, PacketLineType.DELIMITER):
                    self.last_packet = packet
                    return
                yield packet
            except IncompleteReadError as e:
                if len(e.partial) > 0:
                    raise ValueError(f"Unexpected EOF while reading packet length: {e.partial}")
                raise ValueError("Reached EOF before FLUSH packet") from e

    async def _write_packet(self, packet: PacketLine) -> None:
        assert self.writer is not None

        self.writer.write(packet.marker_bytes)
        self.writer.write(packet.data)
        await self.writer.drain()

    async def _read_header_packet(self, section: AsyncIterator[PacketLine]) -> str:
        header = await anext(section, None)
        if header is None:
            assert self.last_packet is not None, "Expected last packet to be set"
            # The section was empty, return empty string as the header
            return ""
        if header.type != PacketLineType.DATA:
            raise ValueError(f"Unexpected packet type: {header.type}: {header.data}")
        return header.data_decoded

    async def _read_v1_server_hello(self) -> tuple[dict[str, str], frozenset[str]]:
        """
        Process the first ref packet received from the server.

        The first ref packet is special as it contains the capabilities of the server.
        """
        assert self.negotiated_protocol == 1

        refs: dict[str, str] = {}

        # Read the capabilities from the server (https://git-scm.com/docs/protocol-v2#_capability_advertisement)
        section = self._read_packets_section()
        version_data = await self._read_header_packet(section)

        if version_data == "":
            if not self.last_packet or self.last_packet.type != PacketLineType.FLUSH:
                raise ValueError("Expected flush packet after empty header")
            logger.debug("No refs received, the repository is empty")
            return refs, frozenset()

        # In V1, the version packet is optional
        if version_data.startswith("version "):
            if version_data != "version 1":
                raise ValueError(f"Expected 'version 1' packet, instead got: {version_data}")
            # read next packet
            first_ref_extended = await self._read_header_packet(section)
        else:
            first_ref_extended = version_data

        first_ref_data, capabilities_list = first_ref_extended.split("\x00", 1)
        first_ref = Ref.from_line(first_ref_data)
        refs[first_ref.name] = first_ref.object_id

        async for packet in section:
            assert packet.type == PacketLineType.DATA
            ref = Ref.from_line(packet.data_decoded)
            if ref.name.endswith("^{}"):
                # Skip the peeled refs
                continue
            refs[ref.name] = ref.object_id

        return refs, frozenset(capabilities_list.split(" "))

    def _add_capability_if_supported(self, in_use: set[str], capability: str) -> bool:
        if capability in self.capabilities:
            in_use.add(capability)
            return True
        return False


class FetchConnection[T: Transport](BaseConnection[T]):
    """
    Git Protocol V2 fetch connection.

    Supported operations:

    - List references via ```#ls_refs()```
    - Fetch objects via ```#fetch_objects()```

    """

    def __init__(self, *, transport: T):
        super().__init__(transport=transport)

        self.git_protocol = 2
        self._negotiated_protocol = None

    @abstractmethod
    async def _open_fetch_service_connection(self) -> tuple[StreamReader, StreamWriter]:
        """
        Connect to the remote service and return the reader and writer.
        """
        pass

    async def __aenter__(self) -> "FetchConnection[T]":
        assert self.reader is None
        assert self.writer is None

        self.reader, self.writer = await self._open_fetch_service_connection()

        if self.negotiated_protocol == 2:
            await self._process_hello_v2()
        elif self.negotiated_protocol == 1:
            await self._process_hello_v1()

        logger.debug(f"Connected with capabilities: {self.capabilities}")

        # We expect the server to send us the capabilities and end with a flush packet
        return self

    async def _process_hello_v2(self):
        """
        Process the hello packet from the server when using Git Protocol V2.
        """
        version_packet = await self._read_packet()
        if version_packet.data.strip() != b"version 2":
            logger.error(f"Expected 'version 2' packet, instead got: {version_packet.data}")
            raise ValueError(f"Failed to open fetch connection to {self.transport.url}")

        # Read the capabilities from the server (https://git-scm.com/docs/protocol-v2#_capability_advertisement)
        advertised_capabilities: set[str] = set()
        async for packet in self._read_packets_until_flush():
            assert packet.type == PacketLineType.DATA
            advertised_capabilities.add(packet.data_decoded)

        self.capabilities = frozenset(advertised_capabilities)

    async def _process_hello_v1(self):
        """
        Process the hello packet from the server when using Git Protocol V1.
        """
        refs, advertised_capabilities = await self._read_v1_server_hello()

        self.capabilities = frozenset(advertised_capabilities)
        self._cached_refs = refs

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        if self.writer:
            logger.debug("Closing writer")
            await self._write_packet(PacketLine.FLUSH)
            self.writer.write_eof()
            await self.writer.drain()
            self.writer.close()
            self.writer = None

        if self.reader:
            logger.debug("Closing reader")
            self.reader.feed_eof()
            self.reader = None

        await self._close_service_connection()

    async def _process_ack_section(self, ack_section: AsyncIterator[PacketLine], missing_objects: set[str]):
        async for ack in ack_section:
            if ack.data == b"nak\n":
                break
            if ack.data.startswith(b"ack\0"):
                obj_id = ack.data_decoded[4:]
                try:
                    missing_objects.remove(obj_id)
                except KeyError:
                    logger.warning(f"Received ACK for unknown object: {obj_id}")
            if ack.data == b"ready\n":
                break

        assert self.last_packet, "Expected last packet to be set"
        if self.last_packet.type == PacketLineType.FLUSH:
            raise ConnectionException("Server negotiation failed. Missing objects: %s" % (missing_objects,))

    async def _generate_command_v2(
        self,
        command: str,
        args: Iterable[str],
        **capabilities: dict[str, str],
    ) -> AsyncIterator[PacketLine]:
        """
        Send a command to the server with the given arguments and capabilities using GitProtocol V2.

        See: https://git-scm.com/docs/gitprotocol-v2#_command_request
        """
        # Send command
        yield PacketLine.data_from_string(f"command={command}")
        # Send capabilities
        for key, value in capabilities.items():
            data = f"{key}={value}" if len(value) > 0 else key
            yield PacketLine.data_from_string(data)

        yield PacketLine.DELIMITER

        # send arguments
        for arg in args:
            yield PacketLine.data_from_string(arg)

        yield PacketLine.FLUSH

    async def _send_packet_transaction(self, packets: AsyncIterator[PacketLine]) -> None:
        """
        Send a series of packets to the server.

        This is a hook method that can be overridden by subclasses to implement custom behavior.
        """
        assert self.writer is not None
        async for packet in packets:
            await self._write_packet(packet)

    async def _send_command_v2(
        self,
        command: str,
        args: Iterable[str],
        **capabilities: dict[str, str],
    ) -> None:
        await self._send_packet_transaction(self._generate_command_v2(command, args, **capabilities))

    async def ls_refs(self, prefix: str = "") -> AsyncIterator[Ref]:
        """
        List references on the remote server.

        For V1, the refs are cached from the advertisement packet.
        For V2: https://git-scm.com/docs/gitprotocol-v2#_ls_refs
        """
        if self.negotiated_protocol == 1:
            assert self._cached_refs is not None

            logger.debug("Using cached refs from advertisement: %s", self._cached_refs)

            # Used cached refs from advertisement
            for name, obj_id in self._cached_refs.items():
                if prefix and not name.startswith(prefix):
                    continue
                yield Ref(name, obj_id)
            return

        assert self.negotiated_protocol == 2

        args: list[str] = []
        if prefix:
            args.append(f"ref-prefix {prefix}")
        # Send the command
        await self._send_command_v2("ls-refs", args)

        # Read the response
        async for packet in self._read_packets_until_flush():
            assert packet.type == PacketLineType.DATA
            ref = Ref.from_line(packet.data_decoded)
            if prefix and not ref.name.startswith(prefix):
                continue
            yield ref

    async def fetch_objects(
        self,
        objects: set[str],
        *,
        have: set[str] | None = None,
        output: AsyncBufferedIOBase,
    ) -> None:
        """
        Fetch objects from the remote server.
        """
        if self.negotiated_protocol == 1:
            return await self._fetch_objects_v1(objects, have=have, output=output)

        if self.negotiated_protocol == 2:
            return await self._fetch_objects_v2(objects, have=have, output=output)

        raise ConnectionException("Unsupported protocol version: %s" % self.negotiated_protocol)

    async def _generate_fetch_v1_request(self, want: set[str], have: set[str] | None) -> AsyncIterator[PacketLine]:
        capabilities: set[str] = set()
        capabilities.add("agent=kalandra")

        self._add_capability_if_supported(capabilities, "multi_ack_detailed")

        # TODO: support side-band in v1
        # if not self._add_capability_if_supported(capabilities, "side-band-64k"):
        #     self._add_capability_if_supported(capabilities, "side-band")

        assert len(want) >= 1, "Expected at least one object to fetch"
        want_iter = iter(want)

        # Send first want + capabilities
        obj = next(want_iter)
        yield PacketLine.data_from_string(f"want {obj} {' '.join(sorted(capabilities))}")

        # Send the rest of the wants
        for obj in want_iter:
            yield PacketLine.data_from_string("want %s" % obj)

        # Send the oids of objects we have
        if have:
            for obj in have:
                # Make sure we don't send NULL_OBJECT_ID by mistake
                if obj == NULL_OBJECT_ID:
                    continue
                yield PacketLine.data_from_string("have %s" % obj)

        # Send done
        yield PacketLine.FLUSH
        yield PacketLine.data_from_string("done")

    async def _fetch_objects_v1(
        self,
        objects: set[str],
        *,
        have: set[str] | None = None,
        output: AsyncBufferedIOBase,
    ) -> None:
        """
        (Git Protocol V1) Fetch objects from the remote server.

        Spec:

        -   https://git-scm.com/docs/http-protocol#_smart_service_git_upload_pack
        -   https://git-scm.com/docs/pack-protocol#_packfile_negotiation
        """
        await self._send_packet_transaction(self._generate_fetch_v1_request(objects, have))

        assert self.reader is not None

        pkt = await self._read_packet()
        while pkt.data.startswith(b"ACK "):
            # we will zero or more ACKs - we don't care about them
            pkt = await self._read_packet()

        if pkt.data.rstrip() != b"NAK":
            raise ConnectionError("Expected NAK packet, instead got: %r" % pkt.data)

        # The rest of the response is the packfile
        async for chunk in self.reader:
            await output.write(chunk)

        await output.flush()

    async def _fetch_objects_v2(
        self,
        objects: set[str],
        *,
        have: set[str] | None = None,
        output: AsyncBufferedIOBase,
    ) -> None:
        """
        (Git Protocol V2) Fetch objects from the remote server.

        Spec: https://git-scm.com/docs/gitprotocol-v2#_fetch
        """
        # Send the command
        base_args = ()
        if "wait-for-done" in self.capabilities:
            base_args += ("wait-for-done",)

        have_args = ("have " + obj for obj in have) if have else ()
        want_args = ("want " + obj for obj in objects)
        await self._send_command_v2("fetch", args=itertools.chain(base_args, have_args, want_args, ("done",)))

        # NOTE: we always send the "done" immediately not waiting for the server to send us the acks
        #       as we can't really do anything with missing objects anyway

        # The response of fetch is broken into a number of sections separated by delimiter packets (0001),
        # with each section beginning with its section header. Most sections are sent only when the packfile is sent.

        # output = acknowledgements flush-pkt |
        #     [acknowledgments delim-pkt] [shallow-info delim-pkt]
        #     [wanted-refs delim-pkt] [packfile-uris delim-pkt]
        #     packfile flush-pkt

        ## 1. acknowledgements flush-pkt | [acknowledgments delim-pkt]
        missing_objects: set[str] = set(objects)

        section = self._read_packets_section()
        header_name = await self._read_header_packet(section)

        if header_name == "acknowledgments":
            await self._process_ack_section(section, missing_objects)
            # after done, missing_objects contains all non-acked objects

            logger.info(f"Did not receive ACKs for {len(missing_objects)} objects")

            # read the next section
            section = self._read_packets_section()
            header_name = await self._read_header_packet(section)

        ## 2. We don't request shallow-info, so we skip it
        ## 3. We don't request wanted-refs, so we skip it
        ## 4. We don't request packfile-uris, so we skip it

        # 5. packfile flush-pkt
        if header_name != "packfile":
            raise ConnectionException(f"Unexpected section: {header_name}")

        async for packet in section:
            assert packet.type == PacketLineType.DATA
            stream_code = packet.data[0]
            if stream_code == 1:
                await output.write(memoryview(packet.data)[1:])
            elif stream_code == 2:
                msg = packet.data_decoded[1:]
                logger.info(msg)
            elif stream_code == 3:
                msg = packet.data_decoded[1:]
                logger.error(msg)

        assert self.last_packet, "Expected last packet to be set"
        if self.last_packet.type != PacketLineType.FLUSH:  # type: ignore
            logger.warning(f"Unexpected packet type at end of packfile: {self.last_packet.type}")


class PushConnection[T: Transport](BaseConnection[T]):
    """
    https://git-scm.com/docs/gitprotocol-pack#_pushing_data_to_a_server
    """

    def __init__(self, *, transport: T):
        super().__init__(transport=transport)

        # git-receive-pack does not support protocol v2 yet, so make sure we use v1
        self.git_protocol = 1
        self.refs = {}

    @abstractmethod
    async def _open_push_service_connection(self) -> tuple[StreamReader, StreamWriter]:
        """
        Connect to the remote service and return the reader and writer.
        """
        pass

    async def __aenter__(self) -> "PushConnection[T]":
        assert self.reader is None
        assert self.writer is None

        self.reader, self.writer = await self._open_push_service_connection()

        # Read the first ref packet, which is special as it will contain the capabilities
        self.refs, self.capabilities = await self._read_v1_server_hello()

        logger.debug(f"Connected with capabilities: {self.capabilities}")

        # We expect the server to send us the capabilities and end with a flush packet
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        if self.reader:
            logger.debug("Closing reader")
            self.reader.feed_eof()
            self.reader = None

        if self.writer:
            logger.debug("Closing writer")
            try:
                await self._write_packet(PacketLine.FLUSH)
                self.writer.write_eof()
                self.writer.close()
            except ConnectionResetError:
                logger.debug("Connection was reset")
            self.writer = None

        await self._close_service_connection()

    async def _generate_receive_commands(
        self,
        changes: list[RefChange],
        supports_delete: bool,
        use_capabilties: set[str],
    ) -> AsyncIterator[PacketLine]:
        first = True

        for change in changes:
            if change.is_delete:
                if not supports_delete:
                    logger.warning(f"Server does not support delete-refs capability, skipping delete of {change.ref}")
                    continue

            line = f"{change.old} {change.new} {change.ref}"
            if first:
                line += "\0 " + " ".join(use_capabilties)
                first = False
            yield PacketLine.data_from_string(line)

        # We need to send flush packet at the end of commands. THIS IS MISSING FROM DOCS!
        yield PacketLine.FLUSH

    async def _push_request_v1(
        self, changes: list[RefChange], packfile: AsyncBufferedIOBase | None
    ) -> tuple[frozenset[str], AsyncIterator[PacketLine], AsyncBufferedIOBase | None]:
        """
        Prepare the push request
        """
        use_capabilties: set[str] = set()

        # add any report-status capability if supported
        self._add_capability_if_supported(use_capabilties, "report-status")
        self._add_capability_if_supported(use_capabilties, "side-band-64k")
        self._add_capability_if_supported(use_capabilties, "object-format=sha1")
        use_capabilties.add("agent=git/2.46.00000")
        supports_delete = "delete-refs" in self.capabilities

        packets = self._generate_receive_commands(changes, supports_delete, use_capabilties)
        # has_non_deletes = any(not change.is_delete for change in changes)

        return frozenset(use_capabilties), packets, packfile  # if has_non_deletes else None

    async def _send_packfile(self, packfile: AsyncBufferedIOBase) -> None:
        assert self.writer is not None
        expected_total = await packfile.seek(0, 2)
        await packfile.seek(0)

        bcount = 0
        total = 0
        async for chunk in packfile:
            self.writer.write(chunk)
            bcount += len(chunk)
            if bcount >= (10 * MEGABYTES):
                await self.writer.drain()
                total += bcount
                bcount = 0
                logger.debug("Sent %.1f/%.1f MB", total / MEGABYTES, expected_total / MEGABYTES)

        await self.writer.drain()
        logger.debug("Packfile sent. Waiting for server response")

    async def _send_commands(self, packets: AsyncIterator[PacketLine], packfile: AsyncBufferedIOBase | None) -> None:
        assert self.writer is not None

        async for packet in packets:
            await self._write_packet(packet)

        if packfile is not None:
            await self._send_packfile(packfile)

    def _handle_report_message(self, message: str) -> None | str:
        if message.startswith("unpack ok"):
            logger.info("Unpack OK")
            return

        if message.startswith("ok "):
            # Ref update was successful
            return

        if message.startswith("ng "):
            # Ref update failed
            return message[3:]

    async def push_changes(self, changes: list[RefChange], packfile: AsyncBufferedIOBase | None) -> None:
        """
        (Git Protocol V1) Send changes to the remote server.

        @see: https://git-scm.com/docs/gitprotocol-pack#_pushing_data_to_a_server
        """
        used_capabilities, packets, packfile_to_send = await self._push_request_v1(changes, packfile)
        await self._send_commands(packets, packfile_to_send)

        if "report-status" in used_capabilities:
            logger.info("Reading report-status")
            # Try to read the report-status
            push_errors: list[str] = []

            async for packet in self._read_packets_until_flush():
                assert packet.type == PacketLineType.DATA
                channel = int(packet.data[0])
                if channel == 2:
                    message = packet.data_decoded[1:]
                    logger.info("[err] %s", message)
                if channel == 1:
                    message = PacketLine.from_buffer(packet.data[1:]).data_decoded
                    logger.debug("[out] %s", message)
                    error = self._handle_report_message(message)
                    if error:
                        push_errors.append(error)

            logger.info("Report-status completed")

            if push_errors:
                for error in push_errors:
                    logger.error(error)
                raise ServerError("Push failed with errors. See logs for details")

        logger.info("Push completed")



import logging
import pathlib
from urllib.parse import urlparse

import github

from kalandra.auth.basic import CredentialProvider

logger = logging.getLogger(__name__)


class GitHubAppCredentialProvider(CredentialProvider):
    _installation_id: int | None = None

    def __init__(self, integration: github.GithubIntegration, installation_id: int | None) -> None:
        self._integration = integration
        self._installation_id = installation_id

    async def get_credentials(self, origin: str) -> tuple[str, str] | None:
        if origin != "github.com":
            return None

        if self._installation_id is None:
            return None

        return ("x-access-token", self._integration.get_access_token(self._installation_id).token)


class GithubAPI:
    def __init__(self, app_id: str, private_key: pathlib.Path) -> None:
        self._auth = github.Auth.AppAuth(app_id=app_id, private_key=lambda: private_key.read_bytes())
        self._integration = github.GithubIntegration(auth=self._auth)
        self._org_installations: dict[str, int | None] = {}

    def get_installation_id(self, org: str) -> None | int:
        if not org:
            raise ValueError("No org provided")

        if org not in self._org_installations:
            installation = self._integration.get_org_installation(org)
            self._org_installations[org] = installation.id if installation else None

        return self._org_installations[org]

    def get_org_api(self, org: str) -> github.Github:
        installation_id = self.get_installation_id(org)
        if installation_id is None:
            raise ValueError(f"No installation found for {org}")

        return github.Github(auth=self._auth.get_installation_auth(installation_id))

    async def get_repo_properties(
        self,
        repo_url: str,
    ) -> dict[str, str]:
        url = urlparse(repo_url)

        path_parts = url.path.strip("/").split("/")
        org = path_parts[0]
        repo_name = path_parts[1]
        if repo_name.endswith(".git"):
            repo_name = repo_name[:-4]

        api = self.get_org_api(org)
        repo = api.get_repo(f"{org}/{repo_name}")

        values: dict[str, str] = {}
        for key, value in repo.custom_properties.items():  # type: ignore
            if isinstance(value, str):
                values[key] = value

        return values

    def crendentials_provider_for_org(self, org: str) -> GitHubAppCredentialProvider:
        installation_id = self.get_installation_id(org)
        return GitHubAppCredentialProvider(self._integration, installation_id)



import asyncio
import logging
from pathlib import Path
from typing import Literal

from kalandra.auth.basic import CredentialProvider

from .base import BaseConnection, FetchConnection, PushConnection, Transport

logger = logging.getLogger(__name__)


class FileConnection(BaseConnection["FileTransport"]):
    process: asyncio.subprocess.Process

    async def _open_service_connection(
        self, service_name: Literal["git-upload-pack", "git-receive-pack"]
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        logger.debug("Connecting to %s", self.transport.path)
        self.process = await asyncio.create_subprocess_exec(
            service_name,
            self.transport.path,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env={"GIT_PROTOCOL": f"version={self.git_protocol}"},
        )

        if self.process.returncode is not None:
            logger.error(f"Failed to start process {service_name}: {self.process} / {self.process.returncode}")
            raise RuntimeError(f"Failed to start process {service_name}")

        assert self.process.stdout is not None
        assert self.process.stdin is not None
        assert self.process.stderr is not None

        # We don't need stderr, so we can just read it in the background
        self._err_task = asyncio.create_task(self.log_error_messages(self.process.stderr))

        return (self.process.stdout, self.process.stdin)

    async def log_error_messages(self, stream: asyncio.StreamReader) -> None:
        async for line in stream:
            logger.error(f"Error: {line.decode()}")

    async def _close_service_connection(self) -> None:
        if self.process.returncode is None:
            self.process.terminate()
        await self.process.wait()

        self._err_task.cancel()
        await self._err_task


class FileFetchConnection(FileConnection, FetchConnection["FileTransport"]):
    async def _open_fetch_service_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-upload-pack")


class FilePushConnection(FileConnection, PushConnection["FileTransport"]):
    async def _open_push_service_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await self._open_service_connection("git-receive-pack")


class FileTransport(Transport):
    def __init__(self, *, url: str, credentials_provider: CredentialProvider):
        assert url.startswith("file://")
        path = Path(url[7:])

        self.path = path.resolve()
        if not self.path.is_dir():
            raise FileNotFoundError(f"Path {self.path} must point to a git repository")

        objects = self.path / "objects"
        if not objects.is_dir():
            raise FileNotFoundError(f"Dir {self.path} doesn't look like a git repository")

        super().__init__(url=path.as_uri(), credentials_provider=credentials_provider)

    @classmethod
    def can_handle_url(cls, url: str) -> bool:
        return url.startswith("file://")

    def fetch(self) -> FileFetchConnection:
        return FileFetchConnection(transport=self)

    def push(self) -> FilePushConnection:
        return FilePushConnection(transport=self)



import netrc
from abc import abstractmethod
from pathlib import Path


class CredentialProvider:
    @abstractmethod
    async def get_credentials(self, origin: str) -> tuple[str, str] | None:
        """
        Get the username and password for the given URL.

        :param origin: The Origin to get the credentials for.
        :return: A tuple of username and password or None if no credentials are available.
        """
        return None  # pragma: no cover


class NoopCredentialProvider(CredentialProvider):
    """
    A credential provider that does nothing.
    """


class NetrcCredentialProvider(CredentialProvider):
    """
    Resolve credentials using the .netrc file.
    """

    def __init__(self, netrc_path: Path | None):
        self._netrc = netrc.netrc(str(netrc_path) if netrc_path else None)

    async def get_credentials(self, origin: str) -> tuple[str, str] | None:
        match = self._netrc.authenticators(origin)
        return (match[0], match[2]) if match else None


class ChainedCredentialProvider(CredentialProvider):
    def __init__(self, *providers: CredentialProvider):
        self._providers = list(providers)

    async def get_credentials(self, origin: str) -> tuple[str, str] | None:
        for provider in self._providers:
            credentials = await provider.get_credentials(origin)
            if credentials is not None:
                return credentials
        return None

    def add_provider(self, provider: CredentialProvider):
        self._providers.append(provider)



import fnmatch
import logging
import re
from typing import AsyncIterator, Callable

import aiofiles

from kalandra.gitprotocol import NULL_OBJECT_ID, Ref, RefChange
from kalandra.transports import Transport

logger = logging.getLogger(__name__)


async def calculate_mirror_updates(
    mirror_refs: dict[str, str],
    upstream_refs: AsyncIterator[Ref],
    *,
    include_filter: Callable[[str], bool] | None = None,
    exclude_filter: Callable[[str], bool] | None = None,
) -> AsyncIterator[RefChange]:
    """
    Calculate the updates that need to be pushed to the mirror.
    """
    include_filter = include_filter or (lambda _: True)
    exclude_filter = exclude_filter or (lambda _: False)

    refs_to_update = {k: v for k, v in mirror_refs.items() if include_filter(k) and not exclude_filter(k)}

    async for ref in upstream_refs:
        if not include_filter(ref.name):
            refs_to_update.pop(ref.name, None)
            continue

        if exclude_filter(ref.name):
            refs_to_update.pop(ref.name, None)
            continue

        old_id = refs_to_update.pop(ref.name, NULL_OBJECT_ID)
        if old_id != ref.object_id:
            yield RefChange(ref.name, old_id, ref.object_id)
        else:
            logger.debug("Skipping %s, already up-to-date", ref.name)

    # Refs to delete
    for ref, old_id in refs_to_update.items():
        yield RefChange(ref, old_id, NULL_OBJECT_ID)


def create_glob_filter(*globs: str) -> Callable[[str], bool] | None:
    """
    Create a filter that matches the given globs.
    """
    if not globs:
        return None

    regex = re.compile("|".join(fnmatch.translate(glob) for glob in globs))

    def filter_func(name: str) -> bool:
        return regex.match(name) is not None

    return filter_func


async def update_mirror(
    upstream: Transport,
    mirror: Transport,
    *,
    dry_run: bool = False,
    include_filter: Callable[[str], bool] | None = None,
    exclude_filter: Callable[[str], bool] | None = None,
) -> list[RefChange]:
    async with (
        upstream.fetch() as upstream_conn,
        mirror.push() as mirror_conn,
    ):
        changes = [
            c
            async for c in calculate_mirror_updates(
                mirror_conn.refs,
                upstream_conn.ls_refs(),
                include_filter=include_filter,
                exclude_filter=exclude_filter,
            )
        ]

        if not changes:
            logger.info("No changes detected")
            return changes

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Following changes detected:")
            for change in changes:
                logger.debug(change)

        if logger.isEnabledFor(logging.INFO):
            logger.info(
                "Following changes detected: %s creates, %d updates, %d deletes",
                sum(c.is_create for c in changes),
                sum(c.is_update for c in changes),
                sum(c.is_delete for c in changes),
            )

        if dry_run:
            return changes

        # Fetch objects from upstream
        async with aiofiles.tempfile.NamedTemporaryFile(suffix=".pack", encoding=None) as packfile:
            new_objects = {change.new for change in changes}
            new_objects.discard(NULL_OBJECT_ID)
            have_objects = set(mirror_conn.refs.values())

            if new_objects:
                logger.info("Fetching objects from upstream")
                await upstream_conn.fetch_objects(new_objects, have=have_objects, output=packfile)
            else:
                logger.info("No new objects to fetch, only deletes or updates")

            # Make sure all objects are written to disk
            await packfile.flush()

            await packfile.seek(0, 2)
            packfile_size = await packfile.tell()
            logger.info("Downloaded packfile size: %.2f MB", packfile_size / 1024 / 1024)

            # Push objects to mirror
            logger.info("Sending changes to mirror")
            await mirror_conn.push_changes(changes, packfile)

        return changes
